/*
Copyright 2006, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2006
*/

#include <ScatterWeb.System.h>
#include <ScatterWeb.Sd.h>
#include <fat.h>

#define LED_ON			P5OUT &=~0x80
#define LED_OFF			P5OUT |= 0x80
#define LED_TOGGLE		P5OUT ^= 0x80

char					buffer[512];
sd_cache_t				cache;

struct {
	char Name[8];
	file_handle_t hFile;
} pathtrack[8];
int pathtrack_index;

uint8_t	global_ret;
uint16_t	global_r2;

fat_state_t fat;

bool Process_radioHandler(struct netpacket_handler_args* args)
{
	/*
		Notes: * The received packet data becomes invalid when this handler is completed.
	*/

	// Pass packet through command interface
	if ( Cmd_radioHandler( args ) ) {
		// packet handled by messaging module
		return true;
	}

	if ( args->cpaResult == ADRCOMP_OTHER ) {
		// intercept packets for other nodes here
	} else if ( args->cpaResult == ADRCOMP_BROADCAST ) {
		// process broadcast packets here
	} else {
		// process packets for this node here
		switch ( args->netheader->type ) {

			default:
				return false;		// packet not handled (unknown)
		}
		return true;				// packet handled
	}
	return false;					// packet not handled (other destination)
}

void Process_sensorHandler(void* sdata) {

}

// memory info
COMMAND(sdm, 0, cmdargs) {
	uint32_t		size = 0;
	uint16_t		mb, kb;

	// read size from csd register
	size = Sd_get_size();

	// calculate card size in megabytes from bytes
	size >>= 10;
	kb = size % 1024;
	size >>= 10;
	mb = size;

	sprintf(cmdargs->response_args, "Supp. blocklength: 2^%u-2^%u size: %u.%u MiB", sd_state.MinBlockLen_bit, sd_state.MaxBlockLen_bit, mb, kb);
}

// manufacturer info
COMMAND(sdi, 0, cmdargs) {
	sd_cid_t	cid;
	
	if( Sd_read_cid(&cid) ) {
		snprintf(cmdargs->response_args, cmdargs->response_bufsize,
			"CID mid:%#.2x, oid:%.2s, pnm:%.5s, psn:%lu, rev:%#.2x, mdt:%#.4x",
			cid.mid, cid.oid, cid.pnm, cid.psn, cid.revision, cid.mdt);
	}
}

// set blocklength
COMMAND(sdl, 0, cmdargs) {
	uint16_t bl = String_parseUint16(cmdargs->args, NULL);

	if( bl > sd_state.MaxBlockLen_bit || bl < sd_state.MinBlockLen_bit ) {
		snprintf(cmdargs->response_args, cmdargs->response_bufsize,
		"ERROR: Supp. blocklength 2^%u-2^%u", sd_state.MinBlockLen_bit, sd_state.MaxBlockLen_bit);
	} else {
		bl = Sd_set_blocklength(bl);
		snprintf(cmdargs->response_args, cmdargs->response_bufsize, "blocklength: 2^%u", sd_state.BlockLen_bit);
	}
}

// read block
COMMAND(sdr, CMDFLAG_SERIAL, cmdargs) {
	uint32_t address = strtoul(cmdargs->args, NULL, 0);
	uint16_t ret;

	ret = Sd_read_block(buffer, address);
	if( ret ) {
		uint8_t	cs[2];
		uint16_t	i;

		printf("%u Bytes follow:\r\n", sd_state.BlockLen);
		for(i = 0; i < sd_state.BlockLen; ) {
			String_fromByteToHex(buffer[i], cs);
			printf("%.2s", cs);
			i++;
			if( (i % 2) == 0 ) printf(" ");
			if( (i % 16) == 0 ) printf("\r\n");
		}
	} else {
		printf("ERROR: Address not block aligned?");
	}
}

void Fat_Info() {
	uint32_t _Fat_GetFilePos(const fat_state_t* pFS, const file_handle_t* hFile);
	uint32_t	adr;

	printf("\t%u Sectors/Cluster, %u Bytes/Sector\r\n", (1 << fat.SecPerClus_bit), (1 << fat.BytsPerSec_bit));
	printf("\tFAT: at %lu\r\n", fat.FatA);

	adr = _Fat_GetFilePos(&fat, &pathtrack[0].hFile);
	printf("\tRoot Dir: %u entries/%u bytes starting at %#lx\r\n", fat.RootEntCnt, fat.RootDirSz, adr);
	
	adr = fat.FirstClusterSA;
	adr	<<= fat.BytsPerSec_bit;
	printf("\tFirst Sector: %u at %#lx\r\n", fat.FirstClusterSA, adr);
}

void Sd_Initialize() {
	bool ret = 0;

	// initialize card before anything else
	ret = Sd_init_card(&cache);
	
	// result of initialization and card check
	printf("SD Initialized result: %u (> 0 is error)\r\n", ret);
	printf("Type\t'sdm'\t\t for card size\r\n");
	printf("Type\t'sdi'\t\t for card information\r\n");
	printf("Type\t'sdl i'\t\t to set blocklength to 2^i\r\n");
	printf("Type\t'sdr a'\t\t to read a block from block aligned address a\r\n");
}

void Fat_Initialize() {
	enum fat_init_result ret;
	
	ret = Fat_Init( &fat );
	printf("FAT Initialized result: %u (> 0 is error)\r\n", ret);
	if( ret == FAT_ERROR_NONE ) {
		Fat_GetRootDir(&fat, &pathtrack[0].hFile);
		pathtrack[0].Name[0] = '\0';
		pathtrack_index = 0;
		Fat_Info();
	}
	
	printf("Type\t'dir'\t\t for directory listing\r\n");
	printf("Type\t'cd dir'\t\t to change directory\r\n");
}

void test_new(int i) {
	file_handle_t	hRoot, hDir, hFile;
	struct fat_direntry  dirEntry;
	unsigned char	ret;

	char			name[13] = "TEST__.DAT";
	fat_date_t		date;
	unsigned int	time;

	Fat_GetRootDir( &fat, &hRoot );
	// /FIRMWARE
	//ret = Fat_FindFileSimple( &fat, &hRoot,					// in 
	//						  dirName, FAT_ATTR_DIRECTORY,	// param
	//						  &dirEntry, &hDir);			// out
	//if( !ret ) return;	

	name[5] = (i % 10) + 0x30;
	name[4] = (i / 10) + 0x30;
	
	Fat_Date(&date, 2007, 01, 03);
	Fat_Time(&time, 11, 49, 59);
	if( !Fat_FileExists(&fat, &hDir, name) ) {
		printf("creating file\r\n");

		ret = Fat_CreateFile(&fat, &hDir, name, 0, date, time, i, &hFile, NULL);

		printf("file created\r\n");
	}
}

void print_pathtrack() {
	int i;
	for( i = 0; i <= pathtrack_index; i++ ) {
		printf( "%.8s\\", &(pathtrack[i].Name[0]) );
	}
	printf("\r\n");
}

COMMAND(fatt, 0, cmdargs) {
	//Fat_Init(&fat);
	test_new(12);
}

COMMAND(dir, CMDFLAG_SERIAL, cmdargs) {
	struct fat_direntry	dirEntry;
	uint32_t		address = 0;
	
	printf("Directory ");
	print_pathtrack();
	printf("Name     Type Attr Size [byte]\r\n");
	File_Reset( &pathtrack[pathtrack_index].hFile );
	while( Fat_GetNextDirEntry( &fat, &pathtrack[pathtrack_index].hFile, &address, &dirEntry ) ) {
		printf("%.8s %.3s  %#.2x %9.lu  \r\n",
			&(dirEntry.Name[0]), &(dirEntry.Name[8]),
			dirEntry.Attributes, dirEntry.FileSize);
	}
}

COMMAND(cd, CMDFLAG_SERIAL, cmdargs) {
	
	if( ((uint8_t*)cmdargs->args)[0] == '.' ) {
		if( ((uint8_t*)cmdargs->args)[1] == '.' ) {
			if( pathtrack_index > 0 )
				pathtrack_index--;
		}
	} else if( ((uint8_t*)cmdargs->args)[0] == '\\' ) {
		pathtrack_index = 0;
	} else if( cmdargs->arg_size > 0 ) {
		
		if( pathtrack_index == 8 ) {
			snprintf(cmdargs->response_args, cmdargs->response_bufsize, 
				"tracking memory full");
			return;
		} else {
			bool result;
			file_handle_t hCurDir;
			memcpy( &hCurDir, &pathtrack[pathtrack_index].hFile, sizeof(file_handle_t));
			
			result = Fat_ChangeDir(&fat, &hCurDir, (char*)cmdargs->args);
			if( result ) {
				// gone down
				pathtrack_index++;
				memcpy( &pathtrack[pathtrack_index].hFile, &hCurDir, sizeof(file_handle_t));
				memcpy( &pathtrack[pathtrack_index].Name, cmdargs->args, 8 );
			} else {
				sprintf(cmdargs->response_args, "failed");
				return;
			}
		}
	}
	int i;
	char* c = cmdargs->response_args;
	for( i = 0; i <= pathtrack_index; i++ ) {
		sprintf( c, "%.8s\\", &(pathtrack[i].Name[0]) );
		c += strlen(&(pathtrack[i].Name[0])) + 1;
	}
}

void Process_init() {
	System_registerCallback(C_RADIO, (fp_vp)Process_radioHandler);
	System_registerCallback(C_SENSOR, (fp_vp)Process_sensorHandler);

	Sd_Initialize();
	Fat_Initialize();
}

// Software header.
APP(
	APPH_STATIC_APP,	// Application Type
	__APP_NAME,			// Application Name (8 chars)
	1, 0				// Aplication Version (2 bytes)
);
