/*
Copyright 2007, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2007
*/

/**
 * @file			ScatterWeb.ADDA.c
 * @ingroup	libADDA
 * @brief			ADDA library
 * 
 * @author		Bastian Blywis
 * @date			Mi Okt 31 11:59:35 2007
 * @since		1.2
 * @version	$Revision$
 * 
 */

#include "ScatterWeb.ADDA.internals.h"
#include "ScatterWeb.ADDA.h"
#include "ScatterWeb.DAC.internals.h"

/******************************************************************************
 * @name	Initialization and configuration
 * @{
 */

void _ADDA_initDAC() {

	/*
	 * DAC12SREF_2: DAC12 select reference voltage = VR+
 	 * DAC12LSEL1: DAC12 latch loads when DAC12_xDAT written (DAC12ENC ignored)
	 * DAC12IR: DAC12 full-scale output = 1x reference voltage
	 * DAC12AM2: Set input and output buffer to low speed and current
	 */
	DAC12_0CTL = DAC12SREF_2 + DAC12LSEL0 + DAC12IR + DAC12AMP_2;
	DAC12_1CTL = DAC12SREF_2 + DAC12LSEL0 + DAC12IR + DAC12AMP_2;
}

bool ADDA_enableDAC() {
	bool on = false;
	if(DAC12_0CTL & DAC12ENC)
		on = true;
	DAC12_0CTL |= DAC12ENC;
	DAC12_1CTL |= DAC12ENC;

	return on;
}

bool ADDA_disableDAC() {
	bool on = false;
	if(DAC12_0CTL & DAC12ENC)
		on = true;
	DAC12_0CTL &= ~DAC12ENC;
	DAC12_1CTL &= ~DAC12ENC;

	return on;
}

void ADDA_initDAC() {
	_ADDA_initDAC();
	ADDA_enableDAC();
	
	LOG_MEDIUM("DAC initialized");
}

/**
 * @brief Set DAC output to given voltage:  <code>setDAC [DAC] [voltage]</code>
 * @ingroup commands
 * @param DAC out of {0,1}
 * @param voltage out of [0.0 - 3.0]
 * @param approx out of {0,1}
 * @return set voltage in [V]
 * @see ADDA_setDAC
 *
 * Set a DAC output to a given voltage.
 */
COMMAND(setDAC, 0, cmdargs) {
	uint8_t* ptr=0;
	uint16_t dac = String_parseUint16(cmdargs->args, &ptr);
	uint8_t approx = 0;
	float volt = ADDA_parseFloat(ptr+1, ptr);
	if(*ptr != '\0')
		approx = String_parseUint16(ptr+1, &ptr);

	volt = ADDA_setDAC(dac, volt, approx);
	if(volt < 0) {
		printf("ERROR\r\n");
	}
	else {
		printf("DAC(%u) = ", dac);
		ADDA_printFloat(volt);
		printf("V\r\n");
	}
}
/** @} */

float ADDA_setDAC(uint8_t dac, float volt, uint8_t approx) {
	if(volt > 3.0 || volt < 0.0) {
		LOG_MEDIUM("DAC: Voltage out of range.");
		return -1;
	}
	if(volt > ADDA_VeRef) {
		LOG_MEDIUM("DAC: Voltage out of range.");
		return -2;
	}

	// calculate DAC value through successive approximation
	uint16_t out = 0;
	uint16_t mask = 0x800;
	while(mask) {
		out |= mask;
		if(ADDA_bits2Volt(out) > volt)
			out &= ~mask;
		mask = mask >> 1;
	}
	if(approx) {
		while(ADDA_bits2Volt(out) < volt)
			out++;
	}

	switch(dac) {
		case 0:
			DAC0 = out;
			break;
		case 1:
			DAC1= out;
			break;
		default:
			LOG_MEDIUM("DAC: unknown DAC selected");
			return -1;
	}

	return ADDA_bits2Volt(out);
}
