/*
Copyright 2006, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2006
*/

/** \file ScatterWeb.MD2.h
 ** \ingroup libMD2
 ** \brief Methods to handle MD2 checksum generation.
 ** <p>
 ** Copyright (C) 1990-2, RSA Data Security, Inc. Created 1990. All
 ** rights reserved.
 ** <p>
 ** License to copy and use this software is granted for
 ** non-commercial Internet Privacy-Enhanced Mail provided that it is
 ** identified as the "RSA Data Security, Inc. MD2 Message Digest
 ** Algorithm" in all material mentioning or referencing this software
 ** or this function.
 ** <p>
 ** RSA Data Security, Inc. makes no representations concerning either
 ** the merchantability of this software or the suitability of this
 ** software for any particular purpose. It is provided "as is"
 ** without express or implied warranty of any kind.
 ** <p>
 ** These notices must be retained in any copies of any part of this
 ** documentation and/or software.
 ** <p>
 ** - source code taken from reference implementation in RFC 1319
 ** - adapted by Harald Sauff,
 **   Technical University of Hamburg-Harburg (TUHH)
 **
 ** @author        Harald Sauff
 ** @version       $Revision$
 ** @since         1.2
 **/ 

#ifndef ScatterWeb_MD2
#define ScatterWeb_MD2

#include <ScatterWeb.System.h>

/**
 * Permutation of 0..255 constructed from the digits of pi. It
 * gives a "random" nonlinear byte substitution operation.
 */
const uint8_t PI_SUBST[] = {
	41, 46, 67, 201, 162, 216, 124, 1, 61, 54, 84, 161, 236, 240, 6,
	19, 98, 167, 5, 243, 192, 199, 115, 140, 152, 147, 43, 217, 188,
	76, 130, 202, 30, 155, 87, 60, 253, 212, 224, 22, 103, 66, 111, 24,
	138, 23, 229, 18, 190, 78, 196, 214, 218, 158, 222, 73, 160, 251,
	245, 142, 187, 47, 238, 122, 169, 104, 121, 145, 21, 178, 7, 63,
	148, 194, 16, 137, 11, 34, 95, 33, 128, 127, 93, 154, 90, 144, 50,
	39, 53, 62, 204, 231, 191, 247, 151, 3, 255, 25, 48, 179, 72, 165,
	181, 209, 215, 94, 146, 42, 172, 86, 170, 198, 79, 184, 56, 210,
	150, 164, 125, 182, 118, 252, 107, 226, 156, 116, 4, 241, 69, 157,
	112, 89, 100, 113, 135, 32, 134, 91, 207, 101, 230, 45, 168, 2, 27,
	96, 37, 173, 174, 176, 185, 246, 28, 70, 97, 105, 52, 64, 126, 15,
	85, 71, 163, 35, 221, 81, 175, 58, 195, 92, 249, 206, 186, 197,
	234, 38, 44, 83, 13, 110, 133, 40, 132, 9, 211, 223, 205, 244, 65,
	129, 77, 82, 106, 220, 55, 200, 108, 193, 171, 250, 36, 225, 123,
	8, 12, 189, 177, 74, 120, 136, 149, 139, 227, 99, 232, 109, 233,
	203, 213, 254, 59, 0, 29, 57, 242, 239, 183, 14, 102, 88, 208, 228,
	166, 119, 114, 248, 235, 117, 75, 10, 49, 68, 80, 180, 143, 237,
	31, 26, 219, 153, 141, 51, 159, 17, 131, 20
};

/**
 * @brief MD2 context
 *
 * Stores vital data to checksum calculation.
 */
typedef struct {
	uint8_t state[16];			///< state
	uint8_t checksum[16];	///< checksum
	uint16_t count;				///< number of bytes, modulo 16
	uint8_t buffer[16];		///< input buffer
} MD2_CTX;

/**
 * interesting enough: the algorithm runs faster, when memcpy() is
 * called indirectly through MD2_memcpy() - when calling memcpy()
 * directly from the program, you lose about 0.05-0.1ms per hash
 * \todo verify statement
 */
inline void MD2_memcpy(uint8_t* output, uint8_t* input, uint16_t len);

/**
* interesting enough: the algorithm runs faster, when memset() is
* called indirectly through MD2_memset() - when calling memset()
* directly from the program, you lose about 0.05-0.1ms per hash
* \todo verify statement
*/
inline void MD2_memset(uint8_t* output, uint16_t value, uint16_t len);

/**
 * MD2 initialization. Begins an MD2 operation, writing a new context.
 */
inline void MD2_init(MD2_CTX *context);

/**
 * MD2 basic transformation. Transforms state and updates checksum
 * based on block.
 */
void MD2_transform(uint8_t *state, uint8_t *checksum, const uint8_t *block);

/**
 * MD2 block update operation. Continues an MD2 message-digest
 * operation, processing another message block, and updating the
 * context
 */
void MD2_update(MD2_CTX *context, const uint8_t *input, uint16_t inputLen);

/**
 * MD2 finalization. Ends an MD2 message-digest operation, writing the
 * message digest and zeroizing the context.
 */
void MD2_final(uint8_t *digest, MD2_CTX *context);

#endif
