/*
Copyright 2006, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2006
*/

/** \file ScatterWeb.MD4.h
 ** \ingroup libMD4
 ** \brief Methods to handle MD4 checksum generation.
  *
 * <p>
 * Copyright (C) 1991-2, RSA Data Security, Inc. Created 1991. All
 * rights reserved.
 * <p>
 * License to copy and use this software is granted provided that it
 * is identified as the "RSA Data Security, Inc. MD4 Message-Digest
 * Algorithm" in all material mentioning or referencing this software
 * or this function.
 * <p>
 * License is also granted to make and use derivative works provided
 * that such works are identified as "derived from the RSA Data
 * Security, Inc. MD4 Message-Digest Algorithm" in all material
 * mentioning or referencing the derived work.
 * <p>
 * RSA Data Security, Inc. makes no representations concerning either
 * the merchantability of this software or the suitability of this
 * software for any particular purpose. It is provided "as is"
 * without express or implied warranty of any kind.
 * <p>
 * These notices must be retained in any copies of any part of this
 * documentation and/or software.
 * <p>
 * - source code taken from reference implementation in RFC 1320
 * - adapted by Harald Sauff,
 *   Technical University of Hamburg-Harburg (TUHH)
 *
 * @author	Harald Sauff
 * @version	$Revision$
 * @since	1.2
 **/ 

#ifndef ScatterWeb_MD4
#define ScatterWeb_MD4

#include <ScatterWeb.System.h>

// Constants for MD4Transform routine
#define S11 3
#define S12 7
#define S13 11
#define S14 19
#define S21 3
#define S22 5
#define S23 9
#define S24 13
#define S31 3
#define S32 9
#define S33 11
#define S34 15

// F, G and H are basic MD4 functions
#define F(x, y, z) (((x) & (y)) | ((~x) & (z)))
#define G(x, y, z) (((x) & (y)) | ((x) & (z)) | ((y) & (z)))
#define H(x, y, z) ((x) ^ (y) ^ (z))

// ROTATE_LEFT rotates x left n bits
#define ROTATE_LEFT(x, n) (((x) << (n)) | ((x) >> (32-(n))))

// FF, GG and HH are transformations for rounds 1, 2 and 3
// Rotation is separate from addition to prevent recomputation
#define FF(a, b, c, d, x, s) { \
    (a) += F ((b), (c), (d)) + (x); \
    (a) = ROTATE_LEFT ((a), (s)); \
  }
#define GG(a, b, c, d, x, s) { \
    (a) += G ((b), (c), (d)) + (x) + (uint32_t)0x5a827999; \
    (a) = ROTATE_LEFT ((a), (s)); \
  }
#define HH(a, b, c, d, x, s) { \
    (a) += H ((b), (c), (d)) + (x) + (uint32_t)0x6ed9eba1; \
    (a) = ROTATE_LEFT ((a), (s)); \
  }

const uint8_t PADDING[] = {
  0x80, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

/**
 * @brief MD4 context
 *
 * Stores vital data to checksum calculation.
 */
typedef struct {
	uint32_t state[4];	///< state (ABCD)
	uint32_t count[2];	///< number of bits, modulo 2^64 (lsb first)
	uint8_t buffer[64];	///< input buffer
} MD4_CTX;

/**
 * interesting enough: the algorithm runs faster, when memcpy() is
 * called indirectly through MD4_memcpy() - when calling memcpy()
 * directly from the program, you lose about 0.05-0.1ms per hash
 * \todo verify statement
 */
inline void MD4_memcpy(uint8_t* output, uint8_t* input, uint16_t len);

/**
 * interesting enough: the algorithm runs faster, when memcpy() is
 * called indirectly through MD4_memcpy() - when calling memcpy()
 * directly from the program, you lose about 0.05-0.1ms per hash
 * \todo verify statement
 */
inline void MD4_memset(uint8_t* output, uint16_t value, uint16_t len);

/**
 * MD4 initialization. Begins an MD4 operation, writing a new context.
 */
inline void MD4_init(MD4_CTX *context);

/**
 * MD4 basic transformation. Transforms state and updates checksum
 * based on block.
 */
void MD4_transform(uint32_t *state, const uint8_t *block);

/**
 * MD4 block update operation. Continues an MD4 message-digest
 * operation, processing another message block, and updating the
 * context
 */
void MD4_update(MD4_CTX *context, const uint8_t *input, uint16_t inputLen);

/**
 * MD4 finalization. Ends an MD4 message-digest operation, writing the
 * message digest and zeroizing the context.
 */
void MD4_final(uint8_t *digest, MD4_CTX *context);

#endif
